import click
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars
import sys


@click.command("prod-artifacts")
@click.pass_obj
def cli(environment):

    """No-op command: Download your dbt artifacts from production.

    This command is a no-op as the implementation will differ depending on your
    production environment. You should override this command in your own
    project by running `palm override --name prod-artifacts` and then
    implementing the logic to download your dbt artifacts from production.
    """

    plugin_config = environment.plugin_config("dbt")
    artifact_path = plugin_config.dbt_artifacts_prod

    msg = "No-op command! Please override this command in your own project by running: 'palm override --name prod-artifacts'"

    env_vars = dbt_env_vars(environment.palm.branch)
    # success, msg = environment.run_in_docker(cmd, env_vars)

    click.secho(msg, fg="red")
    sys.exit(2)



import re
from typing import Dict
from palm.plugins.dbt.local_user_lookup import local_user_lookup

""" Shared dbt utilities to build out common CLI options """


def dbt_env_vars(branch: str) -> Dict:
    return {
        'PDP_DEV_SCHEMA': _generate_schema_from_branch(branch),
        'PDP_ENV': 'DEVELOPMENT',  # Deprecated - this will be removed!
        'PALM_DBT_ENV': 'DEVELOPMENT',
    }


def _generate_schema_from_branch(branch: str) -> str:
    """Formats the branch name as a schema."""
    user = local_user_lookup()
    return re.sub(r"[^0-9a-zA-Z]+", "_", f"{user}_{branch}").strip("_").lower()



from typing import Optional
import click
from pathlib import Path
from palm.plugins.dbt.dbt_containerizer import DbtContainerizer


@click.command("containerize")
@click.option(
    "--version",
    multiple=False,
    help="dbt version to use (e.g. 1.0.1)",
)
@click.pass_context
def cli(ctx, version: Optional[str]):
    if not version:
        version = click.prompt("Enter dbt version to use", type=str, default="1.0.1")

    template_dir = Path(Path(__file__).parents[1], "templates") / "containerize"
    DbtContainerizer(ctx, template_dir, version).run()
    click.secho(f"Containerized {ctx.obj.palm.image_name}", fg="green")



import click
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command("dbt-docs")
@click.pass_context
def cli(ctx):
    """generates dbt docs and serves them at http://localhost:8080"""
    cmd = "dbt docs generate && dbt docs serve"
    env_vars = dbt_env_vars(ctx.obj.palm.branch)
    success, msg = ctx.obj.run_in_docker(cmd, env_vars)
    click.secho(msg, fg="green" if success else "red")



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars

# TODO: Refactor this command to reduce branching logic and simply join a list of
# commands once, before running. Also use f-strings instead of concat


@click.command("cycle")
@click.argument("count", type=int, default=2)
@click.option(
    "--persist", is_flag=True, help="will not drop the test schema at the end"
)
@click.option("--models", multiple=True, help="see dbt docs on models flag")
@click.option("--select", multiple=True, help="see dbt docs on select flag")
@click.option("--seed", is_flag=True, help="will skip seed full refresh")
@click.pass_context
def cli(
    ctx,
    count: int,
    persist: bool,
    seed: bool,
    models: Optional[Tuple] = tuple(),
    select: Optional[Tuple] = tuple(),
):
    """Consecutive run-test of the dbt repo. `count` is the number of run/test cycles to execute, defaults to 2"""

    def add_models(command: str) -> str:
        if models:
            command += " --models " + " ".join(models)
        return command

    def add_select(command: str) -> str:
        if select:
            command += " --select " + " ".join(select)
        return command

    def run_test():
        return " && ".join(
            (
                add_models("dbt run"),
                add_models("dbt test"),
            )
            * count
        )

    def make_cmd():
        commands = []
        if seed:
            seed_cmd = "dbt seed --full-refresh"
            commands.append(add_select(seed_cmd))

        commands.append(run_test())

        if not persist:
            commands.append("dbt run-operation drop_branch_schemas")

        return " && ".join(list(filter(None, commands)))

    env_vars = dbt_env_vars(ctx.obj.palm.branch)
    cmd = make_cmd()
    success, msg = ctx.obj.run_in_docker(cmd, env_vars)
    click.secho(msg, fg="green" if success else "red")



from typing import Optional
from pathlib import Path
import yaml
from .project_parser import Project


def parse_project(project_path: Optional[Path] = Path("dbt_project.yml")) -> Project:
    """Parse the dbt project file

    Args:
        project_path (Path): Path to the dbt project file

    Returns:
        Project: Parsed project file
    """

    if not project_path.exists():
        raise FileNotFoundError(f"Could not find {project_path}")

    config = yaml.safe_load(project_path.read_text())
    project = Project(**config)
    return project



import click
import subprocess
from pathlib import Path
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command("cleanup")
@click.pass_context
def cli(ctx):
    """Removes any artifacts from Snowflake related to the current branch."""

    cmd = "dbt run-operation drop_branch_schemas && dbt clean && dbt deps"
    env_vars = dbt_env_vars(ctx.obj.palm.branch)
    success, msg = ctx.obj.run_in_docker(cmd, env_vars)
    click.secho(msg, fg="green" if success else "red")

    click.echo("Remote cleanup complete! Cleaning your local docker env...")
    subprocess.run("docker-compose down", check=True, shell=True, cwd=Path.cwd())
    click.echo("Congratulations, you are squeaky clean!")



import click
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command('shell')
@click.pass_context
def cli(ctx):
    """Bash into your docker image to run arbitrary commands"""
    env_vars = dbt_env_vars(ctx.obj.palm.branch)
    ctx.obj.run_in_docker('bash', env_vars)



class Project:
    """Project parser model"""

    def __init__(self, **data):
        self.name = data['name']
        self.version = data['version']
        self.profile = data['profile']
        self.config_version = data['config-version']
        self.model_paths = data.get('model-paths', 'models')
        self.macro_paths = data.get('macro-paths', 'macros')
        self.seed_paths = data.get('seed-paths', 'seeds')
        self.snapshot_paths = data.get('snapshot-paths', 'snapshots')
        self.analysis_paths = data.get('analysis-paths', 'analysis')
        self.test_paths = data.get('test-paths', 'tests')
        self.packages_install_path = data.get('packages-install-path', 'packages')
        # Supports dbt < 1.0.0 - this will be removed in an upcoming release
        self.modules_path = data.get('modules-path', 'dbt_modules')
        self.docs_paths = data.get('docs-paths', [])
        self.vars = data.get('vars', {})
        self.seeds = data.get('seeds', {})
        self.models = data.get('models', {})
        self.snapshots = data.get('snapshots', {})



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command("snapshot")
@click.option("--clean", is_flag=True, help="Drop the test schema after the run")
@click.option("--select", "-s", multiple=True, help="See dbt docs on select flag")
@click.option("--exclude", "-e", multiple=True, help="See dbt docs on exclude flag")
@click.pass_obj
def cli(
    environment,
    clean: bool,
    select: Optional[Tuple] = tuple(),
    exclude: Optional[Tuple] = tuple(),
):
    """Executes the dbt snapshots."""
    env_vars = dbt_env_vars(environment.palm.branch)

    cmd = ["dbt snapshot"]
    if select:
        cmd.append("--select")
        cmd.extend(select)
    if exclude:
        cmd.append('--exclude')
        cmd.extend(exclude)

    success, msg = environment.run_in_docker(" ".join(cmd), env_vars)
    click.secho(msg, fg="green" if success else "red")

    if clean:
        success, msg = environment.run_in_docker(
            "dbt run-operation drop_branch_schemas", env_vars
        )
        click.secho(msg, fg="green" if success else "red")



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command("compile")
@click.option("--models", "-m", multiple=True, help="See dbt docs on models flag")
@click.option("--select", "-s", multiple=True, help="See dbt docs on select flag")
@click.option("--exclude", "-e", multiple=True, help="See dbt docs on exclude flag")
@click.pass_obj
def cli(
    environment,
    models: Optional[Tuple] = tuple(),
    select: Optional[Tuple] = tuple(),
    exclude: Optional[Tuple] = tuple(),
):
    """Compiles the dbt repo"""
    env_vars = dbt_env_vars(environment.palm.branch)

    # --select and --models are interchangeable on dbt >= v1, combine the lists of selections
    targets = list(set(models + select))

    cmd = ["dbt compile"]
    if targets:
        cmd.append("--select")
        cmd.extend(targets)
    if exclude:
        cmd.append('--exclude')
        cmd.extend(exclude)

    success, msg = environment.run_in_docker(" ".join(cmd), env_vars)
    click.secho(msg, fg="green" if success else "red")



import click
import re
from typing import Optional
from pathlib import Path
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars
import palm.plugins.dbt.sql_to_dbt as sql_to_dbt


valid_model_types = ['tmp', 'staging', 'intermediate', 'dim', 'fact']


@click.group()
def cli():
    """dbt model tools"""
    pass


@cli.command('new')
@click.pass_context
@click.option(
    "--name",
    multiple=False,
    required=True,
    help="Name of the model you are creating. Don't add 'dim_' or 'fact_' we do that for you!",
)
@click.option(
    "--model-type",
    multiple=False,
    default='dim',
    help="staging, intermediate, fact, dim - default is dim",
)
@click.option(
    "--model-dir",
    multiple=False,
    help="Subdirectory in /models directory for your model",
)
@click.option(
    "--use-ref-file",
    is_flag=True,
    help="Parses .palm/model_template/ref_files/ref_file.sql into .sql and .yml templates",
)
def new(
    ctx,
    name: str,
    model_type: str,
    model_dir: Optional[str] = '',
    use_ref_file: Optional[bool] = False,
):
    """Generate a new model

    Generates a new model in your application's /models directory.
    """
    click.echo(f'Generating your new {name} {model_type}!')

    if model_type not in valid_model_types:
        click.secho(
            f'Invalid model type. Valid model types are: {", ".join(valid_model_types)}',
            fg='red',
        )
        return

    model_name = get_model_name(name, model_type)

    if use_ref_file:
        ref_file = sql_to_dbt.get_ref_file()
        # Check that ref file exists
        if len(ref_file) == 0:
            click.secho("Your project does not have a ref file", fg="red")
            if click.confirm("Set up a ref file?"):
                sql_to_dbt.create_ref_files()
            return
        # Check that ref file has been updated
        if re.match('^[<].*[>]$', ref_file):
            click.secho(
                "Ref file not updated, please add your source SQL and re-run the command",
                fg="red",
            )
            return

    create_model(model_name, model_dir, model_type, ctx, use_ref_file)
    create_docs(model_name, model_dir, model_type, ctx, use_ref_file)

    # if use_ref_file:
    #     env_vars = dbt_env_vars(ctx.obj.palm.branch)
    #     success, msg = ctx.obj.run_in_docker(f"dbt run --models @{model_name} --fail-fast", env_vars)
    #     click.secho(msg, fg="green" if success else "red")


def create_model(model_name, model_dir, model_type, ctx, use_ref_file):
    model_template_path = Path(Path(__file__).parent.parent, 'model_template', 'model')
    models_path = Path('models/', model_dir, model_type)

    replacements = {
        'model_name': model_name,
        'model_description': f"'{{{{ doc(\"{model_name}\") }}}}'",
        'ref': "{{ ref('staging_table') }}",
        'model_columns': f"""\n      - name: _sk\n        description: The unique id for {model_name}\n        tests:\n          - not_null""",
    }

    if use_ref_file:
        replacements['model_columns'] = sql_to_dbt.sql_to_yml()

    result = ctx.obj.generate(model_template_path, models_path, replacements)
    click.echo(result)

    if use_ref_file:
        sql_to_dbt.create_dbt_sql_file(model_name, models_path)


def create_docs(model_name, model_dir, model_type, ctx, use_ref_file):
    docs_template_path = Path(Path(__file__).parent.parent, 'model_template', 'docs')
    docs_path = Path('models/', model_dir, '/documentation/models/', model_type)

    replacements = {
        'model_name': model_name,
        'begin_docs': f"{{% docs {model_name} %}}",
        'end_docs': "{% enddocs %}",
        'model_notes': "## Business Notes\n\n\t## Developer Notes",
    }

    if use_ref_file:
        replacements['model_notes'] = sql_to_dbt.sql_to_md()

    result = ctx.obj.generate(docs_template_path, docs_path, replacements)

    click.echo(result)


def get_model_name(name, model_type):
    if model_type == 'dim':
        return f"dim_{name}"
    if model_type == 'fact':
        return f"fact_{name}"
    if model_type == 'tmp':
        return f"tmp_{name}"
    return name



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars
import sys


@click.command('test')
@click.option(
    "--clean", is_flag=True, help="Drop the test schema after the run is complete"
)
@click.option("--models", "-m", multiple=True, help="See dbt docs on models flag")
@click.option("--select", "-s", multiple=True, help="See dbt docs on select flag")
@click.option("--exclude", "-e", multiple=True, help="See dbt docs on exclude flag")
@click.option("--defer", "-d", is_flag=True, help="See dbt docs on defer flag")
@click.option(
    "--no-fail-fast", "-nx", is_flag=True, help="Runs all tests even if one fails"
)
@click.pass_obj
def cli(
    environment,
    clean: bool,
    no_fail_fast: bool,
    defer: bool,
    models: Optional[Tuple] = tuple(),
    select: Optional[Tuple] = tuple(),
    exclude: Optional[Tuple] = tuple(),
):
    """Tests the dbt repo"""

    if defer:
        click.secho("Running 'palm prod-artifacts'...", fg='yellow')
        exit_code, _, _ = environment.run_on_host("palm prod-artifacts")
        if exit_code == 2:
            click.secho(
                "'palm prod-artifacts' not implemented. Can't use --defer without it!",
                fg='red',
            )
            sys.exit(1)
        elif exit_code != 0:
            click.secho(
                "Something went wrong while pulling the prod artifacts.", fg='red'
            )
            sys.exit(1)

    env_vars = set_env_vars(environment, defer)

    # --select and --models are interchangeable on dbt >= v1, combine the lists of selections
    targets = list(set(models + select))

    run_cmd = build_test_command(
        no_fail_fast=no_fail_fast,
        targets=targets,
        exclude=exclude,
        defer=defer,
    )
    success, msg = environment.run_in_docker(run_cmd, env_vars)
    click.secho(msg, fg="green" if success else "red")

    if clean:
        success, msg = environment.run_in_docker(
            "dbt run-operation drop_branch_schemas", env_vars
        )
        click.secho(msg, fg="green" if success else "red")


def build_test_command(
    defer: bool = False,
    no_fail_fast: bool = False,
    targets: Optional[list] = None,
    exclude: Optional[Tuple] = None,
) -> str:
    cmd = []

    cmd.append(f"dbt test")
    if targets:
        cmd.append('--select')
        cmd.extend(targets)
    if exclude:
        cmd.append('--exclude')
        cmd.extend(exclude)
    if not no_fail_fast:
        cmd.append('--fail-fast')
    if defer:
        if not targets:
            cmd.extend(["--select", "state:new", "state:modified+"])
        cmd.append("--defer")

    return " ".join(cmd)


def set_env_vars(environment, defer: bool = False) -> dict:
    plugin_config = environment.plugin_config('dbt')
    env_vars = dbt_env_vars(environment.palm.branch)
    if defer:
        env_vars['DBT_DEFER_TO_STATE'] = 'true'
        env_vars['DBT_ARTIFACT_STATE_PATH'] = plugin_config.dbt_artifacts_prod
    else:
        env_vars['DBT_ARTIFACT_STATE_PATH'] = plugin_config.dbt_artifacts_local
    return env_vars



import click

from typing import Optional
from pathlib import Path
from pydantic import BaseModel
from palm.plugins.base_plugin_config import BasePluginConfig


class dbtPluginConfigModel(BaseModel):
    dbt_artifacts_prod: Optional[str]
    dbt_artifacts_local: str


class DbtPluginConfig(BasePluginConfig):
    def __init__(self):
        super().__init__('dbt', dbtPluginConfigModel)

    def set(self) -> dict:
        config = {}

        has_prod_artifacts = click.confirm(
            "Do you have production dbt artifacts saved locally?"
        )
        if has_prod_artifacts:
            prod = click.prompt(
                "Prod artifacts location:", type=click.Path(exists=True)
            )
            config["dbt_artifacts_prod"] = str(prod)
            click.secho(f"Saved prod artifacts location:  {prod}", fg="green")

        local = click.prompt(
            "Local artifacts location:",
            type=click.Path(exists=True),
            default=Path("target/"),
        )
        config['dbt_artifacts_local'] = str(local)

        return config



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command('seed')
@click.option(
    "--clean", is_flag=True, help="drop the test schema after the run is complete"
)
@click.option("--select", '-s', multiple=True, help="see dbt docs on select flag")
@click.option("--exclude", '-e', multiple=True, help="see dbt docs on exclude flag")
@click.option(
    "--no-full-refresh",
    "-nf",
    is_flag=True,
    help="Insert seeds instead of recreating the table",
)
@click.pass_obj
def cli(
    environment,
    clean: bool,
    no_full_refresh: bool,
    select: Optional[Tuple] = tuple(),
    exclude: Optional[Tuple] = tuple(),
):
    """Run dbt seeds"""

    cmd = ['dbt', 'seed']
    if select:
        cmd.append('--select')
        cmd.extend(select)
    if exclude:
        cmd.append('--exclude')
        cmd.extend(exclude)
    if not no_full_refresh:
        cmd.append('--full-refresh')
    if clean:
        cmd.append('&& dbt run-operation drop_branch_schemas')

    env_vars = dbt_env_vars(environment.palm.branch)
    success, msg = environment.run_in_docker(" ".join(cmd), env_vars)
    click.secho(msg, fg="green" if success else "red")



import pkg_resources
from pathlib import Path

from palm.plugins.dbt.plugin_config import DbtPluginConfig
from palm.plugins.base import BasePlugin


def get_version():
    try:
        version = pkg_resources.require("palm-dbt")[0].version
    except pkg_resources.DistributionNotFound:
        version = 'unknown'
    return version


DbtPlugin = BasePlugin(
    name='dbt',
    command_dir=Path(__file__).parent / 'commands',
    version=get_version(),
    package_location='https://github.com/palmetto/palm-dbt.git',
    config=DbtPluginConfig(),
)



import click
import shutil
from typing import List
from pathlib import Path


@click.command("install")
@click.pass_context
def cli(ctx):
    """Install required palm dbt macros in the dbt project

    Palm uses temporary schemas to ensure running and testing dbt is idempotent
    between changesets. The macros installed in this command enable that functionality
    within any dbt project.

    generate_branch_schemas - creates a schema name for the current git branch. This
    macro _requires_ the PALM_DBT_ENV environment variable be set to DEVELOPMENT | CI | PROD

    drop_branch_schemas - Used to clean up after each run, unless the --persist
    flag is provided. Only runs against a TEST database. Drops schemas matching
    the current git branch name.

    """

    macro_template_path = Path(Path(__file__).parent.parent, "macros")
    macros_path = Path.cwd() / "macros"
    macros = ['drop_branch_schemas.sql', 'generate_schema_name.sql']

    if macros_are_installed(macros_path, macros):
        click.secho("It looks like you already installed palm-dbt macros!", fg="green")
        return

    if not macros_path.exists():
        macros_path.mkdir()

    # Note: we are not using ctx.obj.generate for this because generating jinja marcros
    # is not easy nor fun. Simple file copying is preferable!
    for macro in macros:
        shutil.copy(Path(macro_template_path, macro), Path(macros_path, macro))

    click.secho("Palm dbt macros installed!", fg="green")


def macros_are_installed(macros_path: Path, macros: List[str]) -> bool:
    if not macros_path.exists():
        return False

    for macro in macros:
        if Path(macros_path / macro).exists():
            return True

    return False



from distutils.core import setup
from pathlib import Path
import sys

# require version of setuptools that supports find_namespace_packages
from setuptools import setup

try:
    from setuptools import find_namespace_packages
except ImportError:
    # the user has a downlevel version of setuptools.
    print('Error: palm requires setuptools v40.1.0 or higher.')
    print(
        'Please upgrade setuptools with "pip install --upgrade setuptools" and try again'
    )
    sys.exit(1)

# User-friendly description from README.md
this_directory = Path(__file__).parent
long_description = Path(this_directory, 'README.md').read_text()

setup(
    name='palm-dbt',
    version='0.7.0',
    description='dbt extension for Palm CLI',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Palmetto - Data & Analytics team',
    author_email='data-analytics-team@palmetto.com',
    url='https://github.com/palmetto/palm-dbt',
    packages=find_namespace_packages(include=['palm', 'palm.*']),
    package_data={'': ['*.md', '*.sql', '*.yaml', '*.yml', '*.txt']},
    install_requires=Path("palm/plugins/dbt/requirements.txt").read_text().splitlines(),
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
    ],
)



import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars
import sys


@click.command("run")
@click.option(
    "--no-fail-fast",
    "-nx",
    is_flag=True,
    help="Turns off --fail-fast. See dbt docs on fail-fast flag.",
)
@click.option("--clean", is_flag=True, help="Drop the test schema after the run")
@click.option("--models", "-m", multiple=True, help="See dbt docs on models flag")
@click.option("--select", "-s", multiple=True, help="See dbt docs on select flag")
@click.option("--exclude", "-e", multiple=True, help="See dbt docs on exclude flag")
@click.option("--defer", "-d", is_flag=True, help="See dbt docs on defer flag")
@click.option("--iterative", "-i", is_flag=True, help="Iterative stateful dbt run")
@click.option(
    "--full-refresh",
    "-f",
    is_flag=True,
    help="Will perform a full refresh on incremental models",
)
@click.option("--seed", is_flag=True, help="Run dbt seed before dbt run")
@click.pass_obj
def cli(
    environment,
    no_fail_fast: bool,
    clean: bool,
    full_refresh: bool,
    iterative: bool,
    defer: bool,
    seed: bool,
    models: Optional[Tuple] = tuple(),
    select: Optional[Tuple] = tuple(),
    exclude: Optional[Tuple] = tuple(),
    vars: Optional[str] = None,
):
    """Runs the dbt repo."""
    stateful = iterative or defer

    if defer:
        click.secho("Running 'palm prod-artifacts'...", fg='yellow')
        exit_code, _, _ = environment.run_on_host("palm prod-artifacts")
        if exit_code == 2:
            click.secho(
                "'palm prod-artifacts' not implemented. Can't use --defer without it!",
                fg='red',
            )
            sys.exit(1)
        elif exit_code != 0:
            click.secho(
                "Something went wrong while pulling the prod artifacts.", fg='red'
            )
            sys.exit(1)

    env_vars = set_env_vars(environment, stateful, defer)

    # --select and --models are interchangeable on dbt >= v1, combine the lists of selections
    targets = list(set(models + select))

    run_cmd = build_run_command(
        # Is there a better way to pass these args in?
        full_refresh=full_refresh,
        seed=seed,
        no_fail_fast=(no_fail_fast or iterative),
        targets=targets,
        exclude=exclude,
        defer=defer,
        vars=vars,
    )

    success, msg = environment.run_in_docker(run_cmd, env_vars)

    if iterative:
        while not success:
            click.secho("dbt run had some issues. Fix, then hit continue.", fg="yellow")
            cycle = click.confirm("Continue?", show_default=True)

            if not cycle:
                break

            env_vars = set_env_vars(environment, stateful)
            stateful_run_cmd = build_run_command(
                targets=["result:error", "result:skipped"],
                seed=False,
                no_fail_fast=iterative,
            )
            success, msg = environment.run_in_docker(stateful_run_cmd, env_vars)

    click.secho(msg, fg="green" if success else "red")

    if clean:
        success, msg = environment.run_in_docker(
            "dbt run-operation drop_branch_schemas", env_vars
        )
        click.secho(msg, fg="green" if success else "red")


def build_run_command(
    full_refresh: bool = False,
    seed: bool = True,
    no_fail_fast: bool = False,
    targets: Optional[list] = None,
    exclude: Optional[Tuple] = None,
    defer: bool = False,
    vars: Optional[str] = None,
) -> str:
    cmd = []
    full_refresh_option = " --full-refresh" if full_refresh else ""

    if seed:
        cmd.append(f"dbt seed --full-refresh")
        cmd.append("&&")

    cmd.append(f"dbt run{full_refresh_option}")
    if targets:
        cmd.append("--select")
        cmd.extend(targets)
    if exclude:
        cmd.append("--exclude")
        cmd.extend(exclude)
    if not no_fail_fast:
        cmd.append("--fail-fast")
    if defer:
        if not targets:
            cmd.extend(["--select", "state:new", "state:modified+"])
        cmd.append("--defer")
    if vars:
        cmd.append(f"--vars '{vars}'")

    return " ".join(cmd)


def set_env_vars(environment, stateful: bool, defer: bool = False) -> dict:
    plugin_config = environment.plugin_config('dbt')
    env_vars = dbt_env_vars(environment.palm.branch)
    if stateful:
        env_vars['DBT_DEFER_TO_STATE'] = 'true'
    if defer:
        env_vars['DBT_ARTIFACT_STATE_PATH'] = plugin_config.dbt_artifacts_prod
    else:
        env_vars['DBT_ARTIFACT_STATE_PATH'] = plugin_config.dbt_artifacts_local
    return env_vars



import os


def local_user_lookup() -> str:
    user = "no_user"
    for name in ('LOGNAME', 'USER', 'LNAME', 'USERNAME'):
        user_found = os.environ.get(name)
        if user_found:
            user = user_found
    return user



import click
import sqlparse
import re
import yaml
from pathlib import Path
from functools import lru_cache
from typing import List

from palm.plugins.dbt.parsers import parse_project


@click.command("model-doc")
@click.argument("model", required=True, type=click.Path(exists=True))
@click.pass_obj
def cli(environment, model):
    """Generates initial baseline model.yml for a given SQL model"""
    model_path = Path(model)
    model_name = model_path.stem
    generate_model_md_file(environment, model_path, model_name)
    generate_yml_file(model_path, model_name)


def generate_model_md_file(environment, model_path: Path, model_name: str) -> Path:
    """Generate the model markdown file

    Args:
        environment (palm.Environment.obj): The Palm environment object
        model_path (Path): The path to the model SQL file
        model_name (str): The name of the model

    Returns:
        Path: The path to the generated model markdown file
    """
    destination = get_md_destination_directory(model_path, model_name)
    click.echo(click.style(f"Generating {model_name}.md in {destination}", fg="green"))

    grain = click.prompt("What is the grain of the model?", type=str)
    description = click.prompt(
        "Please provide a user-facing description for this model", type=str
    )
    replacements = {
        "model_name": model_name,
        "model_name_humanized": model_name.replace("_", " ").title(),
        "grain": grain,
        "description": description,
        "begin_docs": f"{{% docs {model_name} %}}",
        "end_docs": "{% enddocs %}",
    }
    template_dir = Path(Path(__file__).parents[1], "templates") / "model_docs"
    environment.generate(template_dir, destination, replacements)
    return destination / f"{model_name}.md"


def _get_potential_model_doc_directories(model_path: Path) -> List[str]:
    """Infer potential model types from the path of the model

    Args:
        model_path (Path): The path to the model

    Returns:
        List[str]: A list of potential model types
    """
    # TODO: Implement dbt_project.yml parsing to get the docs directory (and other things)
    # This should be read from dbt_project configuration
    project_conf = parse_project()
    model_path_dirs = model_path.parent.parts
    if project_conf.docs_paths:
        model_types = []
        for doc_path in project_conf.docs_paths:
            model_docs_path = Path(doc_path, 'models')
            model_doc_types = {dir.stem for dir in model_docs_path.glob("*")}
            model_type_matches = model_doc_types.intersection(model_path_dirs)
            if model_type_matches:
                model_types.extend(
                    list(map(lambda p: model_docs_path / p, model_type_matches))
                )
        return model_types

    else:
        model_docs_path = Path("models/documentation")
        model_doc_types = {dir.stem for dir in model_docs_path.glob("*")}
        return [
            model_docs_path / model_dir
            for model_dir in model_path_dirs
            if model_dir in model_doc_types
        ]


def get_md_destination_directory(model_path: Path, model_name: str) -> Path:
    """Generate the destination for the models markdown file

    Note that the model must have a parent directory that matches the name of
    and existing group of model docs.
    """
    potential_model_dirs = _get_potential_model_doc_directories(model_path)
    if not potential_model_dirs:
        model_docs_dir = click.prompt(
            "Please enter the directory path for the model doc",
            type=Path,
            default="models/documentation",
        )
    elif len(potential_model_dirs) == 1:
        model_docs_dir = potential_model_dirs[0]
    else:
        model_docs_dir = click.prompt(
            "Multiple directories match the model path. Please enter the directory path for the model doc",
            type=click.Choice(potential_model_dirs),
            default=potential_model_dirs[0],
        )

    if not model_docs_dir:
        click.secho(f"Could not determine directory for {model_name} docs", fg="red")
        raise Exception("Could not determine model type")

    return model_docs_dir


def generate_yml_file(model_path: Path, model_name: str) -> Path:
    """Generate the model yml file"""
    click.echo(click.style(f"Generating {model_name}.yml for {model_name}", fg="green"))
    target_file = model_name + ".yml"
    destination = model_path.parent / target_file
    if destination.exists():
        click.echo(click.style(f"{model_name}.yml already exists", fg="yellow"))
        overwrite = click.confirm("Do you want to overwrite it?", default=False)
        if not overwrite:
            return

    model = {
        "version": 2,
        "models": [
            {
                "name": model_name,
                "description": f'{{{{ doc("{model_name}") }}}}',
                "columns": get_model_columns(model_path),
            }
        ],
    }
    destination.write_text(yaml.dump(model, sort_keys=False, default_flow_style=False))
    click.echo(click.style(f"{model_name}.yml generated at {destination}", fg="green"))
    return destination


def get_model_columns(model_path: Path) -> List[str]:
    """Parse the model SQL file and return a list of columns names

    Args:
        model_path (Path): Path to the model SQL file

    Returns:
        List[str]: List of column names
    """
    raw = Path(model_path).read_text()
    raw = re.sub(
        r"{{[A-Za-z\\n\s()=_,]+}}", "", raw
    ).strip()  # Strip out jinja config blocks
    parsed = sqlparse.parse(raw)

    column_identifiers = []

    for statement in parsed:
        # Unknown is acceptable here because dbt models aren't always proper SQL
        if statement.get_type() in ["SELECT", "UNKNOWN"]:
            column_identifiers = get_column_identifiers(statement)

    column_names = get_column_names(column_identifiers)
    columns = create_column_list(column_names)
    return columns


def get_column_identifiers(
    statement: sqlparse.sql.Statement,
) -> List[sqlparse.sql.Identifier]:
    """Get the column identifiers from a SQL statement"""
    identifiers = []
    for token in statement.tokens:
        if token.ttype is None and type(token) is not sqlparse.sql.Function:
            if type(token) is sqlparse.sql.Identifier:
                # If the token is a CTE, skip it
                if not token_is_cte(token):
                    identifiers.append(token)
            elif type(token) is sqlparse.sql.IdentifierList:
                for child_token in token.get_identifiers():
                    if (
                        child_token.ttype is not sqlparse.tokens.Keyword
                        and not token_is_cte(child_token)
                    ):
                        identifiers.append(child_token)

    if len(identifiers) == 0:
        raise Exception("Could not find column identifiers")
    return identifiers


def token_is_cte(token: sqlparse.sql.Token) -> bool:
    """Check if a given token is a CTE name so we can skip documenting it.

    Args:
        token (sqlparse.sql.Token): The token to check

    Returns:
        bool: True if the token is a CTE name, False otherwise
    """
    _, next_token = token.token_next(1)
    return next_token and next_token.value == "AS"


def get_column_names(column_identifiers: List[sqlparse.sql.Identifier]) -> List[str]:
    """Get the column names from a list of identifiers

    Args:
        column_identifiers (List[sqlparse.sql.Identifier]): List of SQL identifiers

    Returns:
        List[str]: List of column names
    """
    columns = []
    for identifier in column_identifiers:
        columns.append(identifier.get_name())
    return columns


def column_has_existing_doc(column_name: str) -> bool:
    """Check if a column has an existing doc"""
    return column_name in existing_column_docs()


@lru_cache(maxsize=1)
def existing_column_docs() -> List[str]:
    """Get a list of existing column docs

    Cached to prevent globbing the filesystem multiple times

    Returns:
        List[str]: List of existing column doc names
    """
    project_conf = parse_project()
    if project_conf.docs_paths:
        column_docs = []
        for doc_path in project_conf.docs_paths:
            column_docs.extend(Path(doc_path, 'columns').glob('*.md'))
    else:
        column_docs = Path("models/documentation/columns").glob("*.md")

    return [column_doc.stem for column_doc in column_docs]


def create_column_list(column_names: List[str]) -> List[dict]:
    """Create a list of column dictionaries

    Args:
        column_names (List[str]): List of column names

    Returns:
        List[dict]: List of column dictionaries
    """
    columns = []
    columns_without_existing_docs = [
        c for c in column_names if not column_has_existing_doc(c)
    ]

    manual_descriptions = click.confirm(
        f'Do you want to manually enter descriptions for {len(columns_without_existing_docs)} columns?',
        default=False,
    )
    for col_name in column_names:
        description = 'TODO: Add description'
        if column_has_existing_doc(col_name):
            description = f'{{{{ doc("{col_name}") }}}}'
        elif manual_descriptions:
            description = click.prompt(
                f'Description for {col_name}', default=description
            )
            # TODO: Add ability to generate column doc file for manual descriptions

        col_dict = {"name": col_name, "description": description}
        columns.append(col_dict)
    return columns



import os, sys
from pathlib import Path
from typing import Optional, Tuple, Dict
from palm.containerizer import PythonContainerizer
from palm.palm_exceptions import AbortPalm
import click
import yaml


class DbtContainerizer(PythonContainerizer):
    """
    Containerizer for dbt projects.
    """

    def __init__(
        self, ctx, template_dir: Path, dbt_version: Optional[str] = '1.0.1'
    ) -> None:
        self.ctx = ctx
        self.project_name = ctx.obj.palm.image_name
        self.template_dir = template_dir
        self.dbt_version = dbt_version
        self.package_manager = ''

    def run(self) -> None:
        """Runs the containerizer.

        Returns:
            None
        """
        is_valid_version, message = self.validate_dbt_version()
        if not is_valid_version:
            click.secho(message, fg="red")
            sys.exit(1)

        self.package_manager = super().detect_package_manager()
        (
            self.profile_host_path,
            self.profile_volume_path,
        ) = self.determine_profile_strategy(Path.cwd())
        if self.profile_host_path:
            self.write_profile_envs()
        super().generate(self.target_dir, self.replacements)

    def validate_dbt_version(self) -> Tuple[bool, str]:
        """Prompts the user for a DBT version.

        Returns:
            str: DBT version
        """
        semver = self.dbt_version.split(".")
        minimum_version = ['0', '19']
        maximum_version = ['1', '3']

        if semver[0] <= minimum_version[0] and semver[1] < minimum_version[1]:
            return (
                False,
                f'Invalid dbt version, must be > {".".join(minimum_version)}',
            )

        if semver[0] > maximum_version[0] or (
            semver[0] == maximum_version[0] and semver[1] > maximum_version[1]
        ):
            return (
                False,
                f'Invalid dbt version, must be < {".".join(maximum_version)}',
            )

        return (True, f'{self.dbt_version} is valid')

    @property
    def is_dbt_v1(self) -> bool:
        """Returns True if dbt version is 1.x.x"""
        return self.dbt_version.split(".")[0] == '1'

    @property
    def replacements(self) -> Dict:
        """
        Return a dictionary of replacements for the dbt template.
        """

        replacements = dict()
        if self.profile_host_path:
            replacements = {
                "dbt_profile_host": self.profile_host_path,
                "profile_volume_mount": ":".join(
                    (
                        "${DBT_PROFILE_HOST}",
                        self.profile_volume_path,
                    )
                ),
            }
        replacements.update(
            {
                "project_name": self.project_name,
                "package_manager": self.package_manager,
                "dbt_version": self.dbt_version,
                "packages_dir": self.get_packages_dir(),
            }
        )
        return replacements

    def validate_python_version(self) -> bool:
        """Pass through function - the PythonContainerizer handles this functionality.

        Returns:
            True
        """
        return True

    def write_profile_envs(self) -> None:
        """Writes dbt profile envars
        to a new .env, or appends to existing env.
        """
        with Path(".env").open("a") as env_file:
            env_file.write(
                (
                    f"DBT_PROFILE_HOST={self.profile_host_path}\n"
                    f"DBT_PROFILES_DIR={self.profile_volume_path}"
                )
            )

    def get_packages_dir(self) -> str:
        deps_dir = "dbt_packages" if self.is_dbt_v1 else "dbt_modules"
        dbt_confg = self.dbt_project_config()
        if dbt_confg.get('modules-path'):
            deps_dir = dbt_confg['modules-path']
        if dbt_confg.get('packages-install-path'):
            deps_dir = dbt_confg['packages-install-path']
        return deps_dir

    def dbt_project_config(self) -> dict:
        config_path = Path("dbt_project.yml")
        if config_path.exists():
            return yaml.safe_load(config_path.read_text())
        return {}

    @classmethod
    def determine_profile_strategy(cls, project_path: "Path") -> Tuple[str, str]:
        """determines where the on-the-host project
        has been storing the profiles.yml file

        Args:
         project_path: the project root to convert
        Returns:
           the host and container volume mount values
        """
        container_default = "/root/.dbt"
        profile_path = os.getenv("DBT_PROFILES_DIR", None)
        if profile_path:
            profiles_dir = Path(profile_path)
            if not profiles_dir.exists():
                raise AbortPalm("Your host has a non-existant DBT_PROFILES_DIR value!")
            if project_path in profiles_dir.parents:
                return cls._relative_paths(profiles_dir, project_path)
            return str(profiles_dir), container_default
        default_profile_path = Path.home() / ".dbt"
        if not default_profile_path.exists():
            click.secho("No DBT profile found. Skipping.", fg="yellow")
            return None, None
        return str(default_profile_path), container_default

    @classmethod
    def _relative_paths(cls, profiles_dir: str, project_path: str) -> Tuple[str, str]:
        """the relative child path of the given profiles dir
        Args:
         profiles_dir: where is the profile?
         project_path: the root of the project
        Returns:
         relative host and container paths <host_relative_path>, <container_absolute_path>,
        """
        return tuple(
            [
                str(profiles_dir).replace(str(project_path), prefix)
                for prefix in (
                    ".",
                    "/app",
                )
            ]
        )



import re
import click
from typing import Tuple
from shutil import copy
from pathlib import Path
from functools import lru_cache


def create_dbt_sql_file(model_name: str, models_path: Path) -> None:
    """Writes dbt-style code to the model.sql based on SQL provided in the reference file

    Args:
        model_name (str): Name of the model
        model_type (str): Type of the model
    """

    replacements = get_replacements()

    output_file = models_path / model_name / f'{model_name}.sql'

    filedata = get_ref_file()

    filedata = re.sub('"', "", filedata, flags=re.IGNORECASE)

    for pat, repl in replacements:
        filedata = re.sub(pat, repl, filedata, flags=re.IGNORECASE)

    filedata = re.sub('join .*\.([a-z_\.]*)', lower_repl, filedata, flags=re.IGNORECASE)

    with open(output_file, 'w') as file:
        file.write(filedata)
        file.close()

    click.echo(f"Updated datasources in {output_file}'s sql file!")


def sql_to_yml() -> str:
    """Generates column documentation for the model based on ref file SQL notes

    Returns:
        str: Formatted description for each column in the provided SQL
    """
    filedata = get_ref_file()
    results = ''

    select_columns = re.sub(
        '(?s:.*)select\W*distinct\W|[ ]{0,}from[\W.\w]{1,}',
        '',
        filedata,
        flags=re.MULTILINE | re.IGNORECASE,
    )
    select_columns = [
        "".join(tuples).lower()
        for tuples in re.findall(
            '([a-z\_]{1,} -- .*)|([a-z\_]{1,}[ ]{0,}$)',
            select_columns,
            flags=re.MULTILINE | re.IGNORECASE,
        )
    ]

    columns = []

    for col in select_columns:
        columns.append(re.split('[ ]{0,}--[ ]{0,}', col))

    for value in columns:

        name = value[0]

        try:
            description = value[1]
        except IndexError:
            description = 'TBD'

        formatted = (
            "\n      - name: " + name + "\n        description: " + description + "\n"
        )

        results = results + formatted

    return results


def sql_to_md() -> str:
    """Pull the top comment from the ref file to inject into the model_notes md

    Returns:
        str: Content for model docs yml
    """
    filedata = get_ref_file()

    model_notes_present = (
        len(re.findall('\/\*\D([\D]*)\*\/([\D\d]*)', filedata, flags=re.MULTILINE)) > 0
    )

    if model_notes_present:
        results = re.sub(
            '\/\*\D([\D]*)\*\/([\D\d]*)', r'\1', filedata, flags=re.MULTILINE
        )
        results = re.sub('(?!\A)^', ' ' * 4, results, flags=re.MULTILINE)
        return results

    return "## Business Notes\n\n\t## Developer Notes"


@lru_cache(maxsize=1)
def get_ref_file() -> str:
    """Read the ref file, cached to reduce amount of times we need to read the file into memory

    Returns:
        str: ref file contents
    """
    ref_file_path = Path.cwd() / '.palm/model_template/ref_files/ref_file.sql'
    if not ref_file_path.exists():
        return ''
    contents = ref_file_path.read_text()
    return contents


def create_ref_files():
    ref_file_templates = Path(__file__).parent / 'model_template/ref_files'
    ref_file_dir = Path.cwd() / '.palm/model_template/ref_files'
    ref_file_dir.mkdir(parents=True, exist_ok=True)
    files = ['ref_file.sql', 'ref_file_readme.md']
    breakpoint
    for file in files:
        copy(ref_file_templates / file, ref_file_dir / file)
    click.secho(
        "Ref files created! Update the ref file and re-run the command to generate your model",
        fg="green",
    )


## helper functions


def get_replacements() -> list:
    """Get list of replacements

    Returns:
        list: list of replacements
    """

    replacements = []
    directory = Path.cwd() / 'source_systems'

    for file in directory.glob('*.yml'):
        replacement_for_file = get_replacements_for_file(file)
        if replacement_for_file:
            replacements.append(replacement_for_file)

    return replacements


def get_replacements_for_file(file: Path) -> Tuple:
    """Extracts the dbt naming conventions from a given file

    Args:
        file ([Path]): Path to a .yml file in sourcE_systems directory

    Returns:
        tuple: path, dbt_path
    """

    with open(file, 'r') as fd:
        sqlFile = fd.read()
        fd.close()

    try:
        src_and_tables = return_regex('src_and_tables', sqlFile)
        database = return_regex('database', sqlFile)[0]
        schema = return_regex('schema', sqlFile)[0]
    except AttributeError:
        print(AttributeError)

    src = src_and_tables[0]

    paths = [database + '.' + schema + '.' + path for path in src_and_tables[1:]]

    for path in paths:
        dbt_path = "{{ source('" + src + "', '" + path.split('.')[2] + "') }}"
        return (path, dbt_path)


def return_regex(typ, txt):

    regex_dict = {
        'src_and_tables': re.compile(
            "- name[: ]{0,3}(.*\_{0,}.*)\n", flags=re.IGNORECASE | re.MULTILINE
        ),
        'database': re.compile(
            "database[: ]{0,3}(.*\_{0,}.*)\n", flags=re.IGNORECASE | re.MULTILINE
        ),
        'schema': re.compile(
            "schema[: ]{0,3}(.*\_{0,}.*)\n", flags=re.IGNORECASE | re.MULTILINE
        ),
    }

    regexp = regex_dict[typ]

    results = regexp.findall(txt)

    results = [result.lower() for result in results]

    if len(results) == 0:
        results.append('ref_only')

    return results


def lower_repl(match):
    return "JOIN {{ ref('" + match.group(1).lower() + "') }}"



from email.policy import default
from unittest.result import failfast
import click
from typing import Optional, Tuple
from palm.plugins.dbt.dbt_palm_utils import dbt_env_vars


@click.command("dbt")
@click.option("--select", '-s', multiple=True, help="Specify the nodes to include.")
@click.option("--exclude", '-e', multiple=True, help="Specify the nodes to exclude.")
@click.option(
    "--selector", help="Specify the selector to use, defined in selectors.yml."
)
@click.option(
    '--fail-fast',
    '-x',
    is_flag=True,
    default=False,
    help='Stop execution upon a first failure',
)
@click.option(
    "--full-refresh", is_flag=True, default=False, help="Specify the nodes to exclude."
)
@click.option(
    "--cleanup",
    is_flag=True,
    default=False,
    help="Drop the schema after running the command.",
)
@click.option(
    '--seed',
    is_flag=True,
    default=False,
    help='Create seeds before running the command.',
)
@click.option(
    "--options",
    '-o',
    help="passthrough for a string of dbt options - useful if an option you need is not available in palm",
)
@click.argument("args", nargs=-1)
@click.pass_context
def cli(
    ctx,
    select: Optional[str],
    exclude: Optional[str],
    selector: Optional[str],
    fail_fast: bool,
    full_refresh: bool,
    seed: bool,
    cleanup: bool,
    options,
    args: Tuple,
):
    """Pass through dbt commands to the container.

    This command allows you to run any dbt command in the Docker container.

    The most frequently used dbt options are available as options to this command.
    For any other options, you can use the --options flag to pass through any
    string of options to dbt.

    Example:
    palm dbt run --select my_model --full-refresh
    palm dbt run --options "--select my_model --full-refresh"
    """
    if len(args) == 0:
        click.secho("You must provide a dbt command", fg="red")
        return

    cmd = [f'dbt {" ".join(args)}']

    if seed:
        cmd.insert(0, f"dbt seed --full-refresh &&")
    if select:
        cmd.append(f" --select {' '.join(select)}")
    if exclude:
        cmd.append(f" --exclude {' '.join(exclude)}")
    if selector:
        cmd.append(f" --selector {selector}")
    if fail_fast:
        cmd.append(" --fail-fast")
    if full_refresh:
        cmd.append(f" --full-refresh")
    if options:
        cmd.append(f" {options}")
    if cleanup:
        cmd.append(" && dbt run-operation drop_branch_schemas")

    env_vars = dbt_env_vars(ctx.obj.palm.branch)
    success, msg = ctx.obj.run_in_docker(" ".join(cmd), env_vars)
    click.secho(msg, fg="green" if success else "red")
